@@ -11,22 +11,22 @@ set -e
 # We need GNU tar as that supports the --mtime and --transform options
 TAR=notfound
 for try in tar gtar gnutar; do
-  if $try -Pcf /dev/null --mtime now /dev/null >/dev/null 2>&1; then
-    TAR=$try
-    break
-  fi
+    if $try -Pcf /dev/null --mtime now /dev/null >/dev/null 2>&1; then
+        TAR=$try
+        break
+    fi
 done
 
 if [ "$TAR" = "notfound" ]; then
-  echo 'No suitable tar (supporting --mtime) found as tar/gtar/gnutar in PATH'
-  exit 1
+    echo 'No suitable tar (supporting --mtime) found as tar/gtar/gnutar in PATH'
+    exit 1
 fi
 
 # Get the current directory, which we'll use for telling Cargo where to find the sources
 wd="$PWD"
 
 # Get the version from git-describe
-VERSION=$(git describe --dirty 2>/dev/null)
+VERSION=$(build_tools/git_version_gen.sh --stdout 2>/dev/null)
 
 # The name of the prefix, which is the directory that you get when you untar
 prefix="fish-$VERSION"
@@ -45,7 +45,7 @@ cd "$PREFIX_TMPDIR"
 mkdir .cargo
 cargo vendor --manifest-path "$wd/Cargo.toml" > .cargo/config.toml
 
-tar cfvJ $path.xz vendor .cargo
+tar cfvJ "$path".xz vendor .cargo
 
 cd -
 rm -r "$PREFIX_TMPDIR"
