@@ -1,11 +1,13 @@
 use super::prelude::*;
+use super::read::TokenOutputMode;
+use crate::ast::{self, Kind, Leaf};
 use crate::common::{unescape_string, UnescapeFlags, UnescapeStringStyle};
 use crate::complete::Completion;
 use crate::expand::{expand_string, ExpandFlags, ExpandResultCode};
 use crate::input::input_function_get_code;
 use crate::input_common::{CharEvent, ReadlineCmd};
 use crate::operation_context::{no_cancel, OperationContext};
-use crate::parse_constants::ParserTestErrorBits;
+use crate::parse_constants::{ParseTreeFlags, ParserTestErrorBits};
 use crate::parse_util::{
     parse_util_detect_errors, parse_util_get_offset_from_line, parse_util_job_extent,
     parse_util_lineno, parse_util_process_extent, parse_util_token_extent,
@@ -22,6 +24,7 @@ use crate::wcstringutil::join_strings;
 use std::ops::Range;
 
 /// Which part of the comandbuffer are we operating on.
+#[derive(Eq, PartialEq)]
 enum TextScope {
     String,
     Job,
@@ -30,21 +33,18 @@ enum TextScope {
 }
 
 /// For text insertion, how should it be done.
+#[derive(Eq, PartialEq)]
 enum AppendMode {
     // replace current text
     Replace,
     // insert at cursor position
     Insert,
+    // insert at cursor position, DWIM style.
+    InsertSmart,
     // insert at end of current token/command/buffer
     Append,
 }
 
-enum TokenMode {
-    Expanded,
-    Raw,
-    Unescaped,
-}
-
 /// Replace/append/insert the selection with/at/after the specified string.
 ///
 /// \param begin beginning of selection
@@ -75,9 +75,12 @@ fn replace_part(
             out.push_utfstr(&buff[range.clone()]);
             out.push_utfstr(insert);
         }
-        AppendMode::Insert => {
+        AppendMode::Insert | AppendMode::InsertSmart => {
             assert!(cursor_pos >= range.start);
             assert!(cursor_pos <= range.end);
+            let insert = strip_dollar_prefixes(insert_mode, &buff[range.start..cursor_pos], insert)
+                .map(Cow::Owned)
+                .unwrap_or(Cow::Borrowed(insert));
             out.push_utfstr(&buff[range.start..cursor_pos]);
             out.push_utfstr(&insert);
             out.push_utfstr(&buff[cursor_pos..range.end]);
@@ -93,6 +96,43 @@ fn replace_part(
     }
 }
 
+// Prefix must be at the beginning of a process.
+fn strip_dollar_prefixes(insert_mode: AppendMode, prefix: &wstr, insert: &wstr) -> Option<WString> {
+    if insert_mode != AppendMode::InsertSmart {
+        return None;
+    }
+    insert.find(L!("$ "))?; // Early return.
+    let source = prefix.to_owned() + insert;
+    let ast = ast::parse(
+        &source,
+        ParseTreeFlags::ACCEPT_INCOMPLETE_TOKENS | ParseTreeFlags::LEAVE_UNTERMINATED,
+        None,
+    );
+    let mut stripped = WString::new();
+    let mut have = prefix.len();
+    for node in ast.walk() {
+        let Kind::DecoratedStatement(ds) = node.kind() else {
+            continue;
+        };
+        let Some(range) = ds.command.range() else {
+            continue;
+        };
+        let pos = range.start();
+        if pos < prefix.len() {
+            continue;
+        }
+        if (pos == 0 || source.as_char_slice()[pos - 1] == '\n')
+            && source.as_char_slice()[pos] == '$'
+            && source.char_at(pos + 1) == ' '
+        {
+            stripped.push_utfstr(&source[have..pos]);
+            have = pos + "$ ".len();
+        }
+    }
+    stripped.push_utfstr(&source[have..]);
+    return Some(stripped);
+}
+
 /// Output the specified selection.
 ///
 /// \param begin start of selection
@@ -104,8 +144,9 @@ fn replace_part(
 fn write_part(
     parser: &Parser,
     range: Range<usize>,
+    range_is_single_token: bool,
     cut_at_cursor: bool,
-    token_mode: Option<TokenMode>,
+    token_mode: Option<TokenOutputMode>,
     buffer: &wstr,
     cursor_pos: usize,
     streams: &mut IoStreams,
@@ -122,21 +163,10 @@ fn write_part(
         return;
     };
 
-    let buff = &buffer[range];
-    let mut tok = Tokenizer::new(buff, TOK_ACCEPT_UNFINISHED);
     let mut args = vec![];
-    while let Some(token) = tok.next() {
-        if cut_at_cursor && token.end() >= pos {
-            break;
-        }
-        if token.type_ != TokenType::string {
-            continue;
-        }
-
-        let token_text = tok.text_of(&token);
-
+    let mut add_token = |token_text: &wstr| {
         match token_mode {
-            TokenMode::Expanded => {
+            TokenOutputMode::Expanded => {
                 const COMMANDLINE_TOKENS_MAX_EXPANSION: usize = 512;
 
                 match expand_string(
@@ -164,10 +194,10 @@ fn write_part(
                     ExpandResultCode::ok => (),
                 };
             }
-            TokenMode::Raw => {
+            TokenOutputMode::Raw => {
                 args.push(Completion::from_completion(token_text.to_owned()));
             }
-            TokenMode::Unescaped => {
+            TokenOutputMode::Unescaped => {
                 let unescaped = unescape_string(
                     token_text,
                     UnescapeStringStyle::Script(UnescapeFlags::INCOMPLETE),
@@ -176,14 +206,39 @@ fn write_part(
                 args.push(Completion::from_completion(unescaped));
             }
         }
-    }
+    };
+
+    let buff = &buffer[range];
+    if range_is_single_token {
+        add_token(buff);
+    } else {
+        let mut tok = Tokenizer::new(buff, TOK_ACCEPT_UNFINISHED);
+        let mut in_redirection = false;
+        while let Some(token) = tok.next() {
+            if cut_at_cursor && token.end() >= pos {
+                break;
+            }
+            let is_redirection_target = in_redirection;
+            in_redirection = token.type_ == TokenType::redirect;
+            if is_redirection_target && token.type_ == TokenType::string {
+                continue;
+            }
+            if token.type_ != TokenType::string {
+                continue;
+            }
+
+            let token_text = tok.text_of(&token);
+            add_token(token_text);
+        }
+    };
+
     for arg in args {
         streams.out.appendln(arg.completion);
     }
 }
 
 /// The commandline builtin. It is used for specifying a new value for the commandline.
-pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> Option<c_int> {
+pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
     let rstate = commandline_get_state(true);
 
     let mut buffer_part = None;
@@ -207,13 +262,13 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
     let mut is_valid = false;
     let mut showing_suggestion = false;
 
-    let mut range = 0..0;
     let mut override_buffer = None;
 
-    const short_options: &wstr = L!(":abijpctfxorhI:CBELSsP");
+    const short_options: &wstr = L!("abijpctfxorhI:CBELSsP");
     let long_options: &[WOption] = &[
         wopt(L!("append"), ArgType::NoArgument, 'a'),
         wopt(L!("insert"), ArgType::NoArgument, 'i'),
+        wopt(L!("insert-smart"), ArgType::NoArgument, '\x06'),
         wopt(L!("replace"), ArgType::NoArgument, 'r'),
         wopt(L!("current-buffer"), ArgType::NoArgument, 'b'),
         wopt(L!("current-job"), ArgType::NoArgument, 'j'),
@@ -247,6 +302,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
             'a' => append_mode = Some(AppendMode::Append),
             'b' => buffer_part = Some(TextScope::String),
             'i' => append_mode = Some(AppendMode::Insert),
+            '\x06' => append_mode = Some(AppendMode::InsertSmart),
             'r' => append_mode = Some(AppendMode::Replace),
             'c' => cut_at_cursor = true,
             't' => buffer_part = Some(TextScope::Token),
@@ -261,12 +317,12 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                         wgettext!("--tokens options are mutually exclusive")
                     ));
                     builtin_print_error_trailer(parser, streams.err, cmd);
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
                 token_mode = Some(match c {
-                    'x' => TokenMode::Expanded,
-                    '\x02' => TokenMode::Raw,
-                    'o' => TokenMode::Unescaped,
+                    'x' => TokenOutputMode::Expanded,
+                    '\x02' => TokenOutputMode::Raw,
+                    'o' => TokenOutputMode::Unescaped,
                     _ => unreachable!(),
                 })
             }
@@ -288,15 +344,19 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
             '\x04' => showing_suggestion = true,
             'h' => {
                 builtin_print_help(parser, streams, cmd);
-                return STATUS_CMD_OK;
+                return Ok(SUCCESS);
             }
             ':' => {
                 builtin_missing_argument(parser, streams, cmd, w.argv[w.wopt_index - 1], true);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
+            }
+            ';' => {
+                builtin_unexpected_argument(parser, streams, cmd, w.argv[w.wopt_index - 1], true);
+                return Err(STATUS_INVALID_ARGS);
             }
             '?' => {
                 builtin_unknown_option(parser, streams, cmd, w.argv[w.wopt_index - 1], true);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             _ => panic!(),
         }
@@ -320,12 +380,12 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
         {
             streams.err.append(wgettext_fmt!(BUILTIN_ERR_COMBO, cmd));
             builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
 
         if positional_args == 0 {
             builtin_missing_argument(parser, streams, cmd, L!("--function"), true);
-            return STATUS_INVALID_ARGS;
+            return Err(STATUS_INVALID_ARGS);
         }
 
         type RL = ReadlineCmd;
@@ -335,7 +395,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                     .err
                     .append(wgettext_fmt!("%ls: Unknown input function '%ls'", cmd, arg));
                 builtin_print_error_trailer(parser, streams.err, cmd);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             };
             // Don't enqueue a repaint if we're currently in the middle of one,
             // because that's an infinite loop.
@@ -349,14 +409,14 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
             reader_execute_readline_cmd(parser, CharEvent::from_readline(cmd));
         }
 
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     if selection_mode {
         if let Some(selection) = rstate.selection {
             streams.out.append(&rstate.text[selection]);
         }
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     // Check for invalid switch combinations.
@@ -365,7 +425,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_TOO_MANY_ARGUMENTS, cmd));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if (search_mode || line_mode || column_mode || cursor_mode || paging_mode)
@@ -375,7 +435,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
             .err
             .append(wgettext_fmt!(BUILTIN_ERR_TOO_MANY_ARGUMENTS, cmd));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if (buffer_part.is_some() || token_mode.is_some() || cut_at_cursor || search_field_mode)
@@ -385,28 +445,28 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
     {
         streams.err.append(wgettext_fmt!(BUILTIN_ERR_COMBO, cmd));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if (token_mode.is_some() || cut_at_cursor) && positional_args != 0 {
         streams.err.append(wgettext_fmt!(
             BUILTIN_ERR_COMBO2,
             cmd,
-            "--cut-at-cursor and --tokens can not be used when setting the commandline"
+            "--cut-at-cursor and token options can not be used when setting the commandline"
         ));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
-    if search_field_mode && buffer_part.is_some() {
-        streams.err.append(wgettext_fmt!(BUILTIN_ERR_COMBO, cmd,));
+    if search_field_mode && (buffer_part.is_some() || token_mode.is_some()) {
+        streams.err.append(wgettext_fmt!(BUILTIN_ERR_COMBO, cmd));
         builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     }
 
     if append_mode.is_some() && positional_args == 0 {
         // No tokens in insert mode just means we do nothing.
-        return STATUS_CMD_ERROR;
+        return Err(STATUS_CMD_ERROR);
     }
 
     // Set default modes.
@@ -414,6 +474,33 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
 
     let buffer_part = buffer_part.unwrap_or(TextScope::String);
 
+    if append_mode == AppendMode::InsertSmart {
+        if search_field_mode {
+            streams.err.append(wgettext_fmt!(
+                BUILTIN_ERR_COMBO2_EXCLUSIVE,
+                cmd,
+                "--insert-smart",
+                "--search-field"
+            ));
+            builtin_print_error_trailer(parser, streams.err, cmd);
+            return Err(STATUS_INVALID_ARGS);
+        }
+        match buffer_part {
+            TextScope::String | TextScope::Job | TextScope::Process => (),
+            TextScope::Token => {
+                // To-do: we can support it in command position.
+                streams.err.append(wgettext_fmt!(
+                    BUILTIN_ERR_COMBO2_EXCLUSIVE,
+                    cmd,
+                    "--insert-smart",
+                    "--current-token"
+                ));
+                builtin_print_error_trailer(parser, streams.err, cmd);
+                return Err(STATUS_INVALID_ARGS);
+            }
+        }
+    }
+
     if line_mode || column_mode {
         if positional_args != 0 {
             let arg = w.argv[w.wopt_index];
@@ -432,7 +519,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                     .err
                     .append(wgettext_fmt!("%ls: line/column index starts at 1", cmd));
                 builtin_print_error_trailer(parser, streams.err, cmd);
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             };
 
             let new_pos = if line_mode {
@@ -444,7 +531,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                         .err
                         .append(wgettext_fmt!("%ls: there is no line %ls\n", cmd, arg));
                     builtin_print_error_trailer(parser, streams.err, cmd);
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 };
                 offset
             } else {
@@ -462,7 +549,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                         arg
                     ));
                     builtin_print_error_trailer(parser, streams.err, cmd);
-                    return STATUS_INVALID_ARGS;
+                    return Err(STATUS_INVALID_ARGS);
                 }
                 line_offset + new_coord
             };
@@ -483,47 +570,47 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                 }
             ));
         }
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     if search_mode {
         return if rstate.search_mode {
-            STATUS_CMD_OK
+            Ok(SUCCESS)
         } else {
-            STATUS_CMD_ERROR
+            Err(STATUS_CMD_ERROR)
         };
     }
 
     if paging_mode {
         return if rstate.pager_mode {
-            STATUS_CMD_OK
+            Ok(SUCCESS)
         } else {
-            STATUS_CMD_ERROR
+            Err(STATUS_CMD_ERROR)
         };
     }
 
     if paging_full_mode {
         return if rstate.pager_mode && rstate.pager_fully_disclosed {
-            STATUS_CMD_OK
+            Ok(SUCCESS)
         } else {
-            STATUS_CMD_ERROR
+            Err(STATUS_CMD_ERROR)
         };
     }
 
     if selection_start_mode {
         let Some(selection) = rstate.selection else {
-            return STATUS_CMD_ERROR;
+            return Err(STATUS_CMD_ERROR);
         };
         streams.out.append(sprintf!("%lu\n", selection.start));
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     if selection_end_mode {
         let Some(selection) = rstate.selection else {
-            return STATUS_CMD_ERROR;
+            return Err(STATUS_CMD_ERROR);
         };
         streams.out.append(sprintf!("%lu\n", selection.end));
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     // At this point we have (nearly) exhausted the options which always operate on the true command
@@ -536,7 +623,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
 
     if search_field_mode {
         let Some((search_field_text, cursor_pos)) = rstate.search_field else {
-            return STATUS_CMD_ERROR;
+            return Err(STATUS_CMD_ERROR);
         };
         transient = search_field_text;
         current_buffer = &transient;
@@ -544,43 +631,35 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
     } else if let Some(override_buffer) = &override_buffer {
         current_buffer = override_buffer;
         current_cursor_pos = current_buffer.len();
-    } else if !parser.libdata().transient_commandlines.is_empty() && !cursor_mode {
-        transient = parser
-            .libdata()
-            .transient_commandlines
-            .last()
-            .unwrap()
-            .clone();
+    } else if parser.libdata().transient_commandline.is_some() && !cursor_mode {
+        transient = parser.libdata().transient_commandline.clone().unwrap();
         current_buffer = &transient;
         current_cursor_pos = transient.len();
-    } else if rstate.initialized {
+    } else if is_interactive_session() {
         current_buffer = &rstate.text;
         current_cursor_pos = rstate.cursor_pos;
     } else {
-        // There is no command line, either because we are not interactive, or because we are
-        // interactive and are still reading init files (in which case we silently ignore this).
-        if !is_interactive_session() {
-            streams.err.append(cmd);
-            streams
-                .err
-                .append(L!(": Can not set commandline in non-interactive mode\n"));
-            builtin_print_error_trailer(parser, streams.err, cmd);
-        }
-        return STATUS_CMD_ERROR;
+        // There is no command line because we are not interactive.
+        streams.err.append(cmd);
+        streams
+            .err
+            .append(L!(": Can not set commandline in non-interactive mode\n"));
+        builtin_print_error_trailer(parser, streams.err, cmd);
+        return Err(STATUS_CMD_ERROR);
     }
 
     if is_valid {
         if current_buffer.is_empty() {
-            return Some(1);
+            return Err(STATUS_CMD_ERROR);
         }
         let res = parse_util_detect_errors(current_buffer, None, /*accept_incomplete=*/ true);
         return match res {
-            Ok(()) => STATUS_CMD_OK,
+            Ok(()) => Ok(SUCCESS),
             Err(err) => {
                 if err.contains(ParserTestErrorBits::INCOMPLETE) {
-                    Some(2)
+                    Err(STATUS_INVALID_ARGS)
                 } else {
-                    STATUS_CMD_ERROR
+                    Err(STATUS_CMD_ERROR)
                 }
             }
         };
@@ -588,11 +667,12 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
 
     if showing_suggestion {
         if reader_showing_suggestion(parser) {
-            return Some(0);
+            return Ok(SUCCESS);
         }
-        return Some(1);
+        return Err(STATUS_CMD_ERROR);
     }
 
+    let range;
     if search_field_mode {
         range = 0..current_buffer.len();
     } else {
@@ -607,7 +687,7 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                 range = parse_util_process_extent(current_buffer, current_cursor_pos, None);
             }
             TextScope::Token => {
-                parse_util_token_extent(current_buffer, current_cursor_pos, &mut range, None);
+                (range, _) = parse_util_token_extent(current_buffer, current_cursor_pos);
             }
         }
     }
@@ -638,35 +718,30 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
                 .out
                 .append(sprintf!("%lu\n", current_cursor_pos - range.start));
         }
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     if positional_args == 0 {
         write_part(
             parser,
             range,
+            buffer_part == TextScope::Token,
             cut_at_cursor,
             token_mode,
             current_buffer,
             current_cursor_pos,
             streams,
         );
-    } else if positional_args == 1 {
-        replace_part(
-            parser,
-            range,
-            args[w.wopt_index],
-            append_mode,
-            current_buffer,
-            current_cursor_pos,
-            search_field_mode,
-        );
     } else {
-        let sb = join_strings(&w.argv[w.wopt_index..], '\n');
+        let replacement = if positional_args == 1 {
+            Cow::Borrowed(args[w.wopt_index])
+        } else {
+            Cow::Owned(join_strings(&w.argv[w.wopt_index..], '\n'))
+        };
         replace_part(
             parser,
             range,
-            &sb,
+            &replacement,
             append_mode,
             current_buffer,
             current_cursor_pos,
@@ -674,5 +749,5 @@ pub fn commandline(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr])
         );
     }
 
-    STATUS_CMD_OK
+    Ok(SUCCESS)
 }
