@@ -2,26 +2,25 @@
 
 use crate::fds::make_fd_blocking;
 use crate::proc::Pid;
-use crate::reader::reader_write_title;
+use crate::reader::{reader_save_screen_state, reader_write_title};
 use crate::tokenizer::tok_command;
 use crate::wutil::perror;
-use crate::{env::EnvMode, proc::TtyTransfer};
+use crate::{env::EnvMode, tty_handoff::TtyHandoff};
 use libc::{STDIN_FILENO, TCSADRAIN};
 
 use super::prelude::*;
 
 /// Builtin for putting a job in the foreground.
-pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Option<c_int> {
-    let opts = match HelpOnlyCmdOpts::parse(argv, parser, streams) {
-        Ok(opts) => opts,
-        Err(err @ Some(_)) if err != STATUS_CMD_OK => return err,
-        Err(err) => panic!("Illogical exit code from parse_options(): {err:?}"),
+pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> BuiltinResult {
+    let opts = HelpOnlyCmdOpts::parse(argv, parser, streams)?;
+
+    let Some(&cmd) = argv.get(0) else {
+        return Err(STATUS_INVALID_ARGS);
     };
 
-    let cmd = argv[0];
     if opts.print_help {
         builtin_print_help(parser, streams, cmd);
-        return STATUS_CMD_OK;
+        return Ok(SUCCESS);
     }
 
     let job;
@@ -40,7 +39,7 @@ pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Optio
                 streams
                     .err
                     .append(wgettext_fmt!("%ls: There are no suitable jobs\n", cmd));
-                return STATUS_INVALID_ARGS;
+                return Err(STATUS_INVALID_ARGS);
             }
             Some((pos, j)) => {
                 job_pos = Some(pos);
@@ -50,7 +49,7 @@ pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Optio
     } else if optind + 1 < argv.len() {
         // Specifying more than one job to put to the foreground is a syntax error, we still
         // try to locate the job $argv[1], since we need to determine which error message to
-        // emit (ambigous job specification vs malformed job id).
+        // emit (ambiguous job specification vs malformed job id).
         let mut found_job = false;
         if let Ok(Some(pid)) = fish_wcstoi(argv[optind]).map(Pid::new) {
             found_job = parser.job_get_from_pid(pid).is_some();
@@ -98,10 +97,7 @@ pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Optio
                     job_pos = Some(pos);
                     job = if !j.wants_job_control() {
                         streams.err.append(wgettext_fmt!(
-                            concat!(
-                                "%ls: Can't put job %d, '%ls' to foreground because ",
-                                "it is not under job control\n"
-                            ),
+                            "%ls: Can't put job %d, '%ls' to foreground because it is not under job control\n",
                             cmd,
                             raw_pid,
                             j.command()
@@ -116,7 +112,7 @@ pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Optio
     };
 
     let Some(job) = job else {
-        return STATUS_INVALID_ARGS;
+        return Err(STATUS_INVALID_ARGS);
     };
     let job_pos = job_pos.unwrap();
 
@@ -143,12 +139,13 @@ pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Optio
 
     // Note if tty transfer fails, we still try running the job.
     parser.job_promote_at(job_pos);
+    let mut handoff = TtyHandoff::new(reader_save_screen_state);
     let _ = make_fd_blocking(STDIN_FILENO);
     {
         let job_group = job.group();
         job_group.set_is_foreground(true);
         if job.entitled_to_terminal() {
-            crate::input_common::terminal_protocols_disable_ifn(false);
+            handoff.disable_tty_protocols();
         }
         let tmodes = job_group.tmodes.borrow();
         if job_group.wants_terminal() && tmodes.is_some() {
@@ -159,19 +156,18 @@ pub fn fg(parser: &Parser, streams: &mut IoStreams, argv: &mut [&wstr]) -> Optio
             }
         }
     }
-    let mut transfer = TtyTransfer::new();
-    transfer.to_job_group(job.group.as_ref().unwrap());
+    handoff.to_job_group(job.group.as_ref().unwrap());
     let resumed = job.resume();
     if resumed {
         job.continue_job(parser);
     }
     if job.is_stopped() {
-        transfer.save_tty_modes();
+        handoff.save_tty_modes();
     }
-    transfer.reclaim();
+    handoff.reclaim();
     if resumed {
-        STATUS_CMD_OK
+        Ok(SUCCESS)
     } else {
-        STATUS_CMD_ERROR
+        Err(STATUS_CMD_ERROR)
     }
 }
