@@ -11,6 +11,7 @@ use crate::global_safety::RelaxedAtomicBool;
 use crate::key;
 use crate::libc::MB_CUR_MAX;
 use crate::parse_util::parse_util_escape_string_with_quote;
+use crate::terminal::Output;
 use crate::termsize::Termsize;
 use crate::wchar::{decode_byte_from_char, encode_byte_to_char, prelude::*};
 use crate::wcstringutil::wcs2string_callback;
@@ -18,9 +19,10 @@ use crate::wildcard::{ANY_CHAR, ANY_STRING, ANY_STRING_RECURSIVE};
 use crate::wutil::encoding::{mbrtowc, wcrtomb, zero_mbstate, AT_LEAST_MB_LEN_MAX};
 use crate::wutil::fish_iswalnum;
 use bitflags::bitflags;
-use core::slice;
-use libc::{EIO, O_WRONLY, SIGTTOU, SIG_IGN, STDERR_FILENO, STDIN_FILENO, STDOUT_FILENO};
+use libc::{SIGTTOU, SIG_IGN, STDIN_FILENO};
 use once_cell::sync::OnceCell;
+use std::cell::{Cell, RefCell};
+use std::env;
 use std::ffi::{CStr, CString, OsStr, OsString};
 use std::mem;
 use std::ops::{Deref, DerefMut};
@@ -29,7 +31,6 @@ use std::path::{Path, PathBuf};
 use std::sync::atomic::{AtomicI32, AtomicU32, Ordering};
 use std::sync::{Arc, MutexGuard};
 use std::time;
-use std::{env, process};
 
 pub const PACKAGE_NAME: &str = env!("CARGO_PKG_NAME");
 
@@ -1068,9 +1069,10 @@ pub static PROGRAM_NAME: OnceCell<&'static wstr> = OnceCell::new();
 /// MS Windows tty devices do not currently have either a read or write timestamp - those respective
 /// fields of `struct stat` are always set to the current time, which means we can't rely on them.
 /// In this case, we assume no external program has written to the terminal behind our back, making
-/// the multiline prompt usable. See #2859 and https://github.com/Microsoft/BashOnWindows/issues/545
+/// the multiline prompt usable. See [#2859](https://github.com/fish-shell/fish-shell/issues/2859)
+/// and <https://github.com/Microsoft/BashOnWindows/issues/545>
 pub fn has_working_tty_timestamps() -> bool {
-    if cfg!(target_os = "windows") {
+    if cfg!(any(target_os = "windows", cygwin)) {
         false
     } else if cfg!(target_os = "linux") {
         !is_windows_subsystem_for_linux(WSL::V1)
@@ -1124,11 +1126,10 @@ pub fn str2wcstring(inp: &[u8]) -> WString {
             // TODO This check used to be conditionally compiled only on affected platforms.
             true
         } else {
-            const _: () = assert!(mem::size_of::<libc::wchar_t>() == mem::size_of::<char>());
             let mut codepoint = u32::from(c);
             ret = unsafe {
                 mbrtowc(
-                    std::ptr::addr_of_mut!(codepoint).cast(),
+                    std::ptr::addr_of_mut!(codepoint),
                     std::ptr::addr_of!(inp[pos]).cast(),
                     inp.len() - pos,
                     &mut state,
@@ -1182,15 +1183,12 @@ pub fn truncate_at_nul(input: &wstr) -> &wstr {
 }
 
 pub fn cstr2wcstring(input: &[u8]) -> WString {
-    let strlen = input.iter().position(|c| *c == b'\0').unwrap();
-    str2wcstring(&input[0..strlen])
+    let input = CStr::from_bytes_until_nul(input).unwrap().to_bytes();
+    str2wcstring(input)
 }
 
 pub(crate) fn charptr2wcstring(input: *const libc::c_char) -> WString {
-    let input: &[u8] = unsafe {
-        let strlen = libc::strlen(input);
-        slice::from_raw_parts(input.cast(), strlen)
-    };
+    let input: &[u8] = unsafe { CStr::from_ptr(input).to_bytes() };
     str2wcstring(input)
 }
 
@@ -1331,16 +1329,14 @@ pub fn fish_setlocale() {
 fn can_be_encoded(wc: char) -> bool {
     let mut converted = [0 as libc::c_char; AT_LEAST_MB_LEN_MAX];
     let mut state = zero_mbstate();
-    unsafe {
-        wcrtomb(converted.as_mut_ptr(), wc as libc::wchar_t, &mut state) != 0_usize.wrapping_sub(1)
-    }
+    unsafe { wcrtomb(converted.as_mut_ptr(), wc as u32, &mut state) != 0_usize.wrapping_sub(1) }
 }
 
 /// Call read, blocking and repeating on EINTR. Exits on EAGAIN.
 /// Return the number of bytes read, or 0 on EOF, or an error.
 pub fn read_blocked(fd: RawFd, buf: &mut [u8]) -> nix::Result<usize> {
     loop {
-        let res = nix::unistd::read(fd, buf);
+        let res = nix::unistd::read(unsafe { BorrowedFd::borrow_raw(fd) }, buf);
         if let Err(nix::Error::EINTR) = res {
             continue;
         }
@@ -1359,9 +1355,7 @@ pub fn valid_func_name(name: &wstr) -> bool {
 
 /// A rusty port of the C++ `write_loop()` function from `common.cpp`. This should be deprecated in
 /// favor of native rust read/write methods at some point.
-///
-/// Returns the number of bytes written or an IO error.
-pub fn write_loop<Fd: AsRawFd>(fd: &Fd, buf: &[u8]) -> std::io::Result<usize> {
+pub fn safe_write_loop<Fd: AsRawFd>(fd: &Fd, buf: &[u8]) -> std::io::Result<()> {
     let fd = fd.as_raw_fd();
     let mut total = 0;
     while total < buf.len() {
@@ -1377,30 +1371,43 @@ pub fn write_loop<Fd: AsRawFd>(fd: &Fd, buf: &[u8]) -> std::io::Result<usize> {
             }
         }
     }
-    Ok(total)
+    Ok(())
 }
 
-/// A rusty port of the C++ `read_loop()` function from `common.cpp`. This should be deprecated in
-/// favor of native rust read/write methods at some point.
-///
-/// Returns the number of bytes read or an IO error.
-pub fn read_loop<Fd: AsRawFd>(fd: &Fd, buf: &mut [u8]) -> std::io::Result<usize> {
-    let fd = fd.as_raw_fd();
-    loop {
-        match nix::unistd::read(fd, buf) {
-            Ok(read) => {
-                return Ok(read);
-            }
-            Err(err) => {
-                if matches!(err, nix::Error::EAGAIN | nix::Error::EINTR) {
-                    continue;
-                }
-                return Err(std::io::Error::from(err));
-            }
-        }
+pub use safe_write_loop as write_loop;
+
+// Output writes always succeed; this adapter allows us to use it in a write-like macro.
+struct OutputWriteAdapter<'a, T: Output>(&'a mut T);
+
+impl<'a, T: Output> std::fmt::Write for OutputWriteAdapter<'a, T> {
+    fn write_str(&mut self, s: &str) -> std::fmt::Result {
+        self.0.write_bytes(s.as_bytes());
+        Ok(())
     }
 }
 
+impl<'a, T: Output> std::io::Write for OutputWriteAdapter<'a, T> {
+    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
+        self.0.write_bytes(buf);
+        Ok(buf.len())
+    }
+    fn flush(&mut self) -> std::io::Result<()> {
+        Ok(())
+    }
+}
+
+pub(crate) fn do_write_to_output(writer: &mut impl Output, args: std::fmt::Arguments<'_>) {
+    let mut adapter = OutputWriteAdapter(writer);
+    std::fmt::write(&mut adapter, args).unwrap()
+}
+
+#[macro_export]
+macro_rules! write_to_output {
+    ($out:expr, $($arg:tt)*) => {{
+        $crate::common::do_write_to_output($out, format_args!($($arg)*));
+    }};
+}
+
 /// Write the given paragraph of output, redoing linebreaks to fit `termsize`.
 pub fn reformat_for_screen(msg: &wstr, termsize: &Termsize) -> WString {
     let mut buff = WString::new();
@@ -1493,7 +1500,7 @@ pub fn restore_term_foreground_process_group_for_exit() {
     // failure because doing so is unlikely to be noticed.
     // Safety: All of getpgrp, signal, and tcsetpgrp are async-signal-safe.
     let initial_fg_process_group = INITIAL_FG_PROCESS_GROUP.load(Ordering::Relaxed);
-    if initial_fg_process_group > 0 && initial_fg_process_group != unsafe { libc::getpgrp() } {
+    if initial_fg_process_group > 0 && initial_fg_process_group != crate::nix::getpgrp() {
         unsafe {
             libc::signal(SIGTTOU, SIG_IGN);
             libc::tcsetpgrp(STDIN_FILENO, initial_fg_process_group);
@@ -1524,7 +1531,7 @@ pub enum WSL {
 /// Determines if we are running under Microsoft's Windows Subsystem for Linux to work around
 /// some known limitations and/or bugs.
 ///
-/// See https://github.com/Microsoft/WSL/issues/423 and Microsoft/WSL#2997
+/// See <https://github.com/Microsoft/WSL/issues/423> and [Microsoft/WSL#2997](https://github.com/Microsoft/WSL/issues/2997)
 #[inline(always)]
 #[cfg(not(target_os = "linux"))]
 pub fn is_windows_subsystem_for_linux(_: WSL) -> bool {
@@ -1534,7 +1541,7 @@ pub fn is_windows_subsystem_for_linux(_: WSL) -> bool {
 /// Determines if we are running under Microsoft's Windows Subsystem for Linux to work around
 /// some known limitations and/or bugs.
 ///
-/// See https://github.com/Microsoft/WSL/issues/423 and Microsoft/WSL#2997
+/// See <https://github.com/Microsoft/WSL/issues/423> and [Microsoft/WSL#2997](https://github.com/Microsoft/WSL/issues/2997)
 #[cfg(target_os = "linux")]
 pub fn is_windows_subsystem_for_linux(v: WSL) -> bool {
     use std::sync::OnceLock;
@@ -1544,7 +1551,7 @@ pub fn is_windows_subsystem_for_linux(v: WSL) -> bool {
     // any allocations or mutexes. We can't rely on all the std functions to be alloc-free in both
     // Debug and Release modes, so we just mandate that the result already be available.
     //
-    // is_wsl() is called by has_working_timestamps() which is called by `screen.cpp` in the main
+    // is_wsl() is called by has_working_tty_timestamps() which is called by `screen.rs` in the main
     // process. If that's not good enough, we can call is_wsl() manually at shell startup.
     if crate::threads::is_forked_child() {
         debug_assert!(
@@ -1554,11 +1561,13 @@ pub fn is_windows_subsystem_for_linux(v: WSL) -> bool {
     }
 
     let wsl = RESULT.get_or_init(|| {
-        let mut info: libc::utsname = unsafe { mem::zeroed() };
-        let release: &[u8] = unsafe {
-            libc::uname(&mut info);
-            std::mem::transmute(&info.release[..])
+        let release = unsafe {
+            let mut info = mem::MaybeUninit::uninit();
+            libc::uname(info.as_mut_ptr());
+            let info = info.assume_init();
+            info.release
         };
+        let release: &[u8] = unsafe { std::mem::transmute(&release[..]) };
 
         // Sample utsname.release under WSLv2, testing for something like `4.19.104-microsoft-standard`
         // or `5.10.16.3-microsoft-standard-WSL2`
@@ -1624,24 +1633,6 @@ pub fn fish_reserved_codepoint(c: char) -> bool {
         || (c >= key::Backspace && c < ENCODE_DIRECT_END)
 }
 
-pub fn redirect_tty_output(in_signal_handler: bool) {
-    unsafe {
-        let mut t: libc::termios = mem::zeroed();
-        let s = CStr::from_bytes_with_nul(b"/dev/null\0").unwrap();
-        let fd = libc::open(s.as_ptr(), O_WRONLY);
-        if in_signal_handler && fd == -1 {
-            process::abort();
-        } else {
-            assert!(fd != -1, "Could not open /dev/null!");
-        }
-        for stdfd in [STDIN_FILENO, STDOUT_FILENO, STDERR_FILENO] {
-            if libc::tcgetattr(stdfd, &mut t) == -1 && errno::errno().0 == EIO {
-                libc::dup2(fd, stdfd);
-            }
-        }
-    }
-}
-
 /// Test if the given char is valid in a variable name.
 pub fn valid_var_name_char(chr: char) -> bool {
     fish_iswalnum(chr) || chr == '_'
@@ -1678,6 +1669,147 @@ pub fn get_executable_path(argv0: impl AsRef<Path>) -> PathBuf {
     argv0.as_ref().to_owned()
 }
 
+/// A wrapper around Cell which supports modifying the contents, scoped to a region of code.
+/// This provides a somewhat nicer API than ScopedRefCell because you can directly modify the value,
+/// instead of requiring an accessor function which returns a mutable reference to a field.
+pub struct ScopedCell<T>(Cell<T>);
+
+impl<T> Deref for ScopedCell<T> {
+    type Target = Cell<T>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+impl<T> DerefMut for ScopedCell<T> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.0
+    }
+}
+
+impl<T: Copy> ScopedCell<T> {
+    pub fn new(value: T) -> Self {
+        Self(Cell::new(value))
+    }
+
+    /// Temporarily modify a value in the ScopedCell, restoring it when the returned object is dropped.
+    ///
+    /// This is useful when you want to apply a change for the duration of a scope
+    /// without having to manually restore the previous value.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use fish::common::ScopedCell;
+    ///
+    /// let cell = ScopedCell::new(5);
+    /// assert_eq!(cell.get(), 5);
+    ///
+    /// {
+    ///     let _guard = cell.scoped_mod(|v| *v += 10);
+    ///     assert_eq!(cell.get(), 15);
+    /// }
+    ///
+    /// // Restored after scope
+    /// assert_eq!(cell.get(), 5);
+    /// ```
+    pub fn scoped_mod<'a, Modifier: FnOnce(&mut T)>(
+        &'a self,
+        modifier: Modifier,
+    ) -> impl ScopeGuarding + 'a {
+        let mut val = self.get();
+        modifier(&mut val);
+        let saved = self.replace(val);
+        ScopeGuard::new(self, move |cell| cell.set(saved))
+    }
+}
+
+/// A wrapper around RefCell which supports modifying the contents, scoped to a region of code.
+pub struct ScopedRefCell<T>(RefCell<T>);
+
+impl<T> Deref for ScopedRefCell<T> {
+    type Target = RefCell<T>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+impl<T> DerefMut for ScopedRefCell<T> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.0
+    }
+}
+
+impl<T> ScopedRefCell<T> {
+    pub fn new(value: T) -> Self {
+        Self(RefCell::new(value))
+    }
+
+    /// Temporarily modify a field in the ScopedRefCell, restoring it when the returned guard is dropped.
+    ///
+    /// This is useful when you want to change part of a data structure for the duration of a scope,
+    /// and automatically restore the original value afterward.
+    ///
+    /// The `accessor` function selects the field to modify by returning a mutable reference to it.
+    ///
+    /// # Example
+    /// ```
+    /// use fish::common::ScopedRefCell;
+    ///
+    /// struct State { flag: bool }
+    ///
+    /// let cell = ScopedRefCell::new(State { flag: false });
+    /// assert_eq!(cell.borrow().flag, false);
+    ///
+    /// {
+    ///     let _guard = cell.scoped_set(true, |s| &mut s.flag);
+    ///     assert_eq!(cell.borrow().flag, true);
+    /// }
+    ///
+    /// // Restored after scope
+    /// assert_eq!(cell.borrow().flag, false);
+    /// ```
+    pub fn scoped_set<'a, Accessor, Value: 'a>(
+        &'a self,
+        value: Value,
+        accessor: Accessor,
+    ) -> impl ScopeGuarding + 'a
+    where
+        Accessor: Fn(&mut T) -> &mut Value + 'a,
+    {
+        let mut data = self.borrow_mut();
+        let mut saved = std::mem::replace(accessor(&mut data), value);
+        ScopeGuard::new(self, move |cell| {
+            let mut data = cell.borrow_mut();
+            std::mem::swap((accessor)(&mut data), &mut saved);
+        })
+    }
+
+    /// Convenience method for replacing the entire contents of the ScopedRefCell, restoring it when dropped.
+    ///
+    /// Equivalent to `scoped_set(value, |s| s)`.
+    ///
+    /// # Example
+    /// ```
+    /// use fish::common::ScopedRefCell;
+    ///
+    /// let cell = ScopedRefCell::new(10);
+    /// assert_eq!(*cell.borrow(), 10);
+    ///
+    /// {
+    ///     let _guard = cell.scoped_replace(99);
+    ///     assert_eq!(*cell.borrow(), 99);
+    /// }
+    ///
+    /// assert_eq!(*cell.borrow(), 10);
+    /// ```
+    pub fn scoped_replace<'a>(&'a self, value: T) -> impl ScopeGuarding + 'a {
+        self.scoped_set(value, |s| s)
+    }
+}
+
 /// A RAII cleanup object. Unlike in C++ where there is no borrow checker, we can't just provide a
 /// callback that modifies live objects willy-nilly because then there would be two &mut references
 /// to the same object - the original variables we keep around to use and their captured references
@@ -1698,26 +1830,24 @@ pub fn get_executable_path(argv0: impl AsRef<Path>) -> PathBuf {
 /// let file = std::fs::File::create("/dev/null").unwrap();
 /// // Create a scope guard to write to the file when the scope expires.
 /// // To be able to still use the file, shadow `file` with the ScopeGuard itself.
-/// let mut file = ScopeGuard::new(file, |file| file.write_all(b"goodbye\n").unwrap());
+/// let mut file = ScopeGuard::new(file, |mut file| file.write_all(b"goodbye\n").unwrap());
 /// // Now write to the file normally "through" the capturing ScopeGuard instance.
 /// file.write_all(b"hello\n").unwrap();
 ///
 /// // hello will be written first, then goodbye.
 /// ```
-pub struct ScopeGuard<T, F: FnOnce(&mut T)>(Option<(T, F)>);
+pub struct ScopeGuard<T, F: FnOnce(T)>(Option<(T, F)>);
 
-impl<T, F: FnOnce(&mut T)> ScopeGuard<T, F> {
+impl<T, F: FnOnce(T)> ScopeGuard<T, F> {
     /// Creates a new `ScopeGuard` wrapping `value`. The `on_drop` callback is executed when the
     /// ScopeGuard's lifetime expires or when it is manually dropped.
     pub fn new(value: T, on_drop: F) -> Self {
         Self(Some((value, on_drop)))
     }
 
-    /// Invokes the callback and returns the wrapped value, consuming the ScopeGuard.
-    pub fn commit(mut guard: Self) -> T {
-        let (mut value, on_drop) = guard.0.take().expect("Should always have Some value");
-        on_drop(&mut value);
-        value
+    /// Invokes the callback, consuming the ScopeGuard.
+    pub fn commit(guard: Self) {
+        std::mem::drop(guard)
     }
 
     /// Cancels the invocation of the callback, returning the original wrapped value.
@@ -1727,7 +1857,7 @@ impl<T, F: FnOnce(&mut T)> ScopeGuard<T, F> {
     }
 }
 
-impl<T, F: FnOnce(&mut T)> Deref for ScopeGuard<T, F> {
+impl<T, F: FnOnce(T)> Deref for ScopeGuard<T, F> {
     type Target = T;
 
     fn deref(&self) -> &Self::Target {
@@ -1735,81 +1865,31 @@ impl<T, F: FnOnce(&mut T)> Deref for ScopeGuard<T, F> {
     }
 }
 
-impl<T, F: FnOnce(&mut T)> DerefMut for ScopeGuard<T, F> {
+impl<T, F: FnOnce(T)> DerefMut for ScopeGuard<T, F> {
     fn deref_mut(&mut self) -> &mut Self::Target {
         &mut self.0.as_mut().unwrap().0
     }
 }
 
-impl<T, F: FnOnce(&mut T)> Drop for ScopeGuard<T, F> {
+impl<T, F: FnOnce(T)> Drop for ScopeGuard<T, F> {
     fn drop(&mut self) {
-        if let Some((mut value, on_drop)) = self.0.take() {
-            on_drop(&mut value);
+        if let Some((value, on_drop)) = self.0.take() {
+            on_drop(value);
         }
     }
 }
 
-/// A trait expressing what ScopeGuard can do. This is necessary because scoped_push returns an
+/// A trait expressing what ScopeGuard can do. This is necessary because our scoped cells return an
 /// `impl Trait` object and therefore methods on ScopeGuard which take a self parameter cannot be
 /// used.
-pub trait ScopeGuarding: DerefMut {
-    /// Invokes the callback and returns the wrapped value, consuming the ScopeGuard.
-    fn commit(guard: Self) -> Self::Target;
-}
-
-impl<T, F: FnOnce(&mut T)> ScopeGuarding for ScopeGuard<T, F> {
-    fn commit(guard: Self) -> T {
-        ScopeGuard::commit(guard)
+pub trait ScopeGuarding: DerefMut + Sized {
+    /// Invokes the callback, consuming the guard.
+    fn commit(guard: Self) {
+        std::mem::drop(guard);
     }
 }
 
-/// A scoped manager to save the current value of some variable, and set it to a new value. When
-/// dropped, it restores the variable to its old value.
-pub fn scoped_push<Context, Accessor, T>(
-    mut ctx: Context,
-    accessor: Accessor,
-    new_value: T,
-) -> impl ScopeGuarding<Target = Context>
-where
-    Accessor: Fn(&mut Context) -> &mut T,
-{
-    let saved = mem::replace(accessor(&mut ctx), new_value);
-    let restore_saved = move |ctx: &mut Context| {
-        *accessor(ctx) = saved;
-    };
-    ScopeGuard::new(ctx, restore_saved)
-}
-
-/// Similar to scoped_push but takes a function like "std::mem::replace" instead of a function
-/// that returns a mutable reference.
-pub fn scoped_push_replacer<Replacer, T>(
-    replacer: Replacer,
-    new_value: T,
-) -> impl ScopeGuarding<Target = ()>
-where
-    Replacer: Fn(T) -> T,
-{
-    let saved = replacer(new_value);
-    let restore_saved = move |_ctx: &mut ()| {
-        replacer(saved);
-    };
-    ScopeGuard::new((), restore_saved)
-}
-
-pub fn scoped_push_replacer_ctx<Context, Replacer, T>(
-    mut ctx: Context,
-    replacer: Replacer,
-    new_value: T,
-) -> impl ScopeGuarding<Target = Context>
-where
-    Replacer: Fn(&mut Context, T) -> T,
-{
-    let saved = replacer(&mut ctx, new_value);
-    let restore_saved = move |ctx: &mut Context| {
-        replacer(ctx, saved);
-    };
-    ScopeGuard::new(ctx, restore_saved)
-}
+impl<T, F: FnOnce(T)> ScopeGuarding for ScopeGuard<T, F> {}
 
 pub const fn assert_send<T: Send>() {}
 pub const fn assert_sync<T: Sync>() {}
@@ -1847,7 +1927,7 @@ pub fn is_console_session() -> bool {
     })
 }
 
-/// Asserts that a slice is alphabetically sorted by a [`&wstr`] `name` field.
+/// Asserts that a slice is alphabetically sorted by a <code>&[wstr]</code> `name` field.
 ///
 /// Mainly useful for static asserts/const eval.
 ///
@@ -1999,3 +2079,17 @@ impl ToCString for &[u8] {
         CString::new(self).unwrap()
     }
 }
+
+#[macro_export]
+macro_rules! env_stack_set_from_env {
+    ($vars:ident, $var_name:literal) => {{
+        use std::os::unix::ffi::OsStrExt;
+        if let Some(var) = std::env::var_os($var_name) {
+            $vars.set_one(
+                L!($var_name),
+                $crate::env::EnvMode::GLOBAL,
+                $crate::common::str2wcstring(var.as_bytes()),
+            );
+        }
+    }};
+}
