@@ -30,6 +30,7 @@ use crate::wildcard::{wildcard_expand_string, wildcard_has_internal};
 use crate::wildcard::{WildcardResult, ANY_CHAR, ANY_STRING, ANY_STRING_RECURSIVE};
 use crate::wutil::{normalize_path, wcstoi_partial, Options};
 use bitflags::bitflags;
+use std::mem::MaybeUninit;
 
 bitflags! {
     /// Set of flags controlling expansions.
@@ -37,8 +38,6 @@ bitflags! {
     pub struct ExpandFlags : u16 {
         /// Fail expansion if there is a command substitution.
         const FAIL_ON_CMDSUBST = 1 << 0;
-        /// Skip command substitutions.
-        const SKIP_CMDSUBST = 1 << 14;
         /// Skip variable expansion.
         const SKIP_VARIABLES = 1 << 1;
         /// Skip wildcard expansion.
@@ -74,6 +73,8 @@ bitflags! {
         const SPECIAL_FOR_COMMAND = 1 << 13;
         /// The token has an unclosed brace, so don't add a space.
         const NO_SPACE_FOR_UNCLOSED_BRACE = 1 << 14;
+        /// Skip command substitutions.
+        const SKIP_CMDSUBST = 1 << 15;
     }
 }
 
@@ -360,7 +361,7 @@ macro_rules! append_syntax_error {
             error.source_start = $source_start;
             error.source_length = 0;
             error.code = ParseErrorCode::syntax;
-            error.text = wgettext_maybe_fmt!($fmt $(, $arg)*);
+            error.text = wgettext_fmt!($fmt $(, $arg)*);
             errors.push(error);
         }
     }
@@ -376,7 +377,7 @@ macro_rules! append_cmdsub_error {
     ) => {
         append_cmdsub_error_formatted!(
             $errors, $source_start, $source_end,
-            wgettext_maybe_fmt!($fmt $(, $arg)*));
+            wgettext_fmt!($fmt $(, $arg)*));
     }
 }
 
@@ -411,7 +412,7 @@ fn append_overflow_error(
         error.text = wgettext!("Expansion produced too many results").to_owned();
         errors.push(error);
     }
-    ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap())
+    ExpandResult::make_error(STATUS_EXPAND_ERROR)
 }
 
 /// Test if the specified string does not contain character which can not be used inside a quoted
@@ -630,11 +631,11 @@ fn expand_variables(
                 errors,
             );
         }
-        return ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap());
+        return ExpandResult::make_error(STATUS_EXPAND_ERROR);
     }
 
     // Do a dirty hack to make sliced history fast (#4650). We expand from either a variable, or a
-    // history_t. Note that "history" is read only in env.cpp so it's safe to special-case it in
+    // history_t. Note that "history" is read only in env.rs so it's safe to special-case it in
     // this way (it cannot be shadowed, etc).
     let mut history = None;
     let mut var = None;
@@ -682,7 +683,7 @@ fn expand_variables(
                         append_syntax_error!(errors, slice_start + bad_pos, "Invalid index value");
                     }
                 }
-                return ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap());
+                return ExpandResult::make_error(STATUS_EXPAND_ERROR);
             }
         }
     }
@@ -860,7 +861,7 @@ fn expand_braces(
 
     if syntax_error {
         append_syntax_error!(errors, SOURCE_LOCATION_UNKNOWN, "Mismatched braces");
-        return ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap());
+        return ExpandResult::make_error(STATUS_EXPAND_ERROR);
     }
 
     let Some(brace_begin) = brace_begin else {
@@ -938,7 +939,7 @@ pub fn expand_cmdsubst(
     ) {
         MaybeParentheses::Error => {
             append_syntax_error!(errors, SOURCE_LOCATION_UNKNOWN, "Mismatched parenthesis");
-            return ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap());
+            return ExpandResult::make_error(STATUS_EXPAND_ERROR);
         }
         MaybeParentheses::None => {
             if !out.add(input) {
@@ -957,31 +958,32 @@ pub fn expand_cmdsubst(
         job_group.as_ref(),
         &mut sub_res,
     );
-    if subshell_status != 0 {
+
+    if let Err(subshell_status) = subshell_status {
         // TODO: Ad-hoc switch, how can we enumerate the possible errors more safely?
         let err = match subshell_status {
-            _ if subshell_status == STATUS_READ_TOO_MUCH.unwrap() => {
+            _ if subshell_status == STATUS_READ_TOO_MUCH => {
                 wgettext!("Too much data emitted by command substitution so it was discarded")
             }
             // TODO: STATUS_CMD_ERROR is overused and too generic. We shouldn't have to test things
             // to figure out what error to show after we've already been given an error code.
-            _ if subshell_status == STATUS_CMD_ERROR.unwrap() => {
+            _ if subshell_status == STATUS_CMD_ERROR => {
                 if ctx.parser().is_eval_depth_exceeded() {
                     wgettext!("Unable to evaluate string substitution")
                 } else {
                     wgettext!("Too many active file descriptors")
                 }
             }
-            _ if subshell_status == STATUS_CMD_UNKNOWN.unwrap() => {
+            _ if subshell_status == STATUS_CMD_UNKNOWN => {
                 wgettext!("Unknown command")
             }
-            _ if subshell_status == STATUS_ILLEGAL_CMD.unwrap() => {
+            _ if subshell_status == STATUS_ILLEGAL_CMD => {
                 wgettext!("Commandname was invalid")
             }
-            _ if subshell_status == STATUS_NOT_EXECUTABLE.unwrap() => {
+            _ if subshell_status == STATUS_NOT_EXECUTABLE => {
                 wgettext!("Command not executable")
             }
-            _ if subshell_status == STATUS_INVALID_ARGS.unwrap() => {
+            _ if subshell_status == STATUS_INVALID_ARGS => {
                 // TODO: Also overused
                 // This is sent for:
                 // invalid redirections or pipes (like `<&foo`),
@@ -990,11 +992,11 @@ pub fn expand_cmdsubst(
                 // time in a background job.
                 wgettext!("Invalid arguments")
             }
-            _ if subshell_status == STATUS_EXPAND_ERROR.unwrap() => {
+            _ if subshell_status == STATUS_EXPAND_ERROR => {
                 // Sent in `for $foo in ...` if $foo expands to more than one word
                 wgettext!("Expansion error")
             }
-            _ if subshell_status == STATUS_UNMATCHED_WILDCARD.unwrap() => {
+            _ if subshell_status == STATUS_UNMATCHED_WILDCARD => {
                 // Sent in `for $foo in ...` if $foo expands to more than one word
                 wgettext!("Unmatched wildcard")
             }
@@ -1026,7 +1028,7 @@ pub fn expand_cmdsubst(
                         append_syntax_error!(errors, slice_begin + bad_pos, "Invalid index value");
                     }
                 }
-                return ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap());
+                return ExpandResult::make_error(STATUS_EXPAND_ERROR);
             }
         };
 
@@ -1156,19 +1158,20 @@ fn expand_home_directory(input: &mut WString, vars: &dyn Environment) {
     } else {
         // Some other user's home directory.
         let name_cstr = wcs2zstring(username);
-        let mut userinfo: libc::passwd = unsafe { std::mem::zeroed() };
+        let mut userinfo = MaybeUninit::uninit();
         let mut result: *mut libc::passwd = std::ptr::null_mut();
         let mut buf = [0 as libc::c_char; 8192];
         let retval = unsafe {
             libc::getpwnam_r(
                 name_cstr.as_ptr(),
-                &mut userinfo,
+                userinfo.as_mut_ptr(),
                 &mut buf[0],
                 std::mem::size_of_val(&buf),
                 &mut result,
             )
         };
         if retval == 0 && !result.is_null() {
+            let userinfo = unsafe { userinfo.assume_init() };
             home = Some(charptr2wcstring(userinfo.pw_dir));
         }
     }
@@ -1181,7 +1184,7 @@ fn expand_home_directory(input: &mut WString, vars: &dyn Environment) {
 /// Expand the %self escape. Note this can only come at the beginning of the string.
 fn expand_percent_self(input: &mut WString) {
     if input.as_char_slice().first() == Some(&PROCESS_EXPAND_SELF) {
-        input.replace_range(0..1, &unsafe { libc::getpid() }.to_wstring());
+        input.replace_range(0..1, &crate::nix::getpid().to_wstring());
     }
 }
 
@@ -1329,7 +1332,7 @@ impl<'a, 'b, 'c> Expander<'a, 'b, 'c> {
             let mut cursor = 0;
             match parse_util_locate_cmdsubst_range(&input, &mut cursor, true, None, None) {
                 MaybeParentheses::Error => {
-                    return ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap());
+                    return ExpandResult::make_error(STATUS_EXPAND_ERROR);
                 }
                 MaybeParentheses::None => {
                     if !out.add(input) {
@@ -1344,7 +1347,7 @@ impl<'a, 'b, 'c> Expander<'a, 'b, 'c> {
                                 parens.end()-1,
                                 "command substitutions not allowed in command position. Try var=(your-cmd) $var ..."
                             );
-                    return ExpandResult::make_error(STATUS_EXPAND_ERROR.unwrap());
+                    return ExpandResult::make_error(STATUS_EXPAND_ERROR);
                 }
             }
         } else {
@@ -1586,6 +1589,5 @@ pub struct ExpandResult {
 
     /// If expansion resulted in an error, this is an appropriate value with which to populate
     /// $status.
-    // todo!("should be c_int?");
-    pub status: i32,
+    pub status: libc::c_int,
 }
