@@ -65,7 +65,7 @@ fn test_error_messages() {
                 string_matches_format(&errors[0].text, fmt),
                 "command '{}' is expected to match error pattern '{}' but is '{}'",
                 $src,
-                $error_text_format,
+                $error_text_format.localize(),
                 &errors[0].text
             );
         };
@@ -102,6 +102,8 @@ fn test_parse_util_process_extent() {
         };
     }
     validate!("for file in (path base\necho", 22, 13..22);
+    validate!("begin\n\n\nec", 10, 6..10);
+    validate!("begin; echo; end", 12, 12..16);
 }
 
 #[test]
