@@ -246,7 +246,7 @@ fn test_convert_private_use() {
         let len = unsafe {
             wcrtomb(
                 std::ptr::addr_of_mut!(converted[0]).cast(),
-                c as libc::wchar_t,
+                c as u32,
                 &mut state,
             )
         };
