@@ -20,18 +20,19 @@ also provides the ability to perform a blocking wait for any topic to change in
 set. This is the real power of topics: you can wait for a sigchld signal OR a thread exit.
 */
 
-use crate::fd_readable_set::FdReadableSet;
+use crate::fd_readable_set::{FdReadableSet, Timeout};
 use crate::fds::{self, make_fd_nonblocking, AutoClosePipes};
 use crate::flog::{FloggableDebug, FLOG};
 use crate::wchar::WString;
 use crate::wutil::perror;
 use nix::errno::Errno;
 use nix::unistd;
-use std::cell::{Cell, UnsafeCell};
+use std::cell::Cell;
 use std::os::fd::AsRawFd;
-use std::pin::Pin;
 use std::sync::atomic::{AtomicU8, Ordering};
 use std::sync::{Condvar, Mutex, MutexGuard};
+#[cfg(target_os = "linux")]
+use std::{cell::UnsafeCell, pin::Pin};
 
 /// The list of topics which may be observed.
 #[repr(u8)]
@@ -159,6 +160,7 @@ impl GenerationsList {
 pub enum BinarySemaphore {
     /// Initialized semaphore.
     /// This is Box'd so it has a stable address.
+    #[cfg(target_os = "linux")]
     Semaphore(Pin<Box<UnsafeCell<libc::sem_t>>>),
     /// Pipes used to emulate a semaphore, if not initialized.
     Pipes(AutoClosePipes),
@@ -198,6 +200,7 @@ impl BinarySemaphore {
     pub fn post(&self) {
         // Beware, we are in a signal handler.
         match self {
+            #[cfg(target_os = "linux")]
             Self::Semaphore(sem) => {
                 let res = unsafe { libc::sem_post(sem.get()) };
                 // sem_post is non-interruptible.
@@ -222,6 +225,7 @@ impl BinarySemaphore {
     /// This loops on EINTR.
     pub fn wait(&self) {
         match self {
+            #[cfg(target_os = "linux")]
             Self::Semaphore(sem) => {
                 loop {
                     match unsafe { libc::sem_wait(sem.get()) } {
@@ -240,10 +244,10 @@ impl BinarySemaphore {
                     // call until data is available (that is, fish would use 100% cpu while waiting for
                     // processes). This call prevents that.
                     if cfg!(feature = "tsan") {
-                        let _ = FdReadableSet::is_fd_readable(fd, FdReadableSet::kNoTimeout);
+                        let _ = FdReadableSet::is_fd_readable(fd, Timeout::Forever);
                     }
                     let mut ignored: u8 = 0;
-                    match unistd::read(fd, std::slice::from_mut(&mut ignored)) {
+                    match unistd::read(&pipes.read, std::slice::from_mut(&mut ignored)) {
                         Ok(1) => break,
                         Ok(_) => continue,
                         // EAGAIN should only be possible if TSAN workarounds have been applied
@@ -261,9 +265,7 @@ impl BinarySemaphore {
     }
 }
 
-// sem_destroy has been deprecated since macOS 10.10 but we only use it under Linux so silence the
-// warning.
-#[cfg_attr(target_os = "macos", allow(deprecated))]
+#[cfg(target_os = "linux")]
 impl Drop for BinarySemaphore {
     fn drop(&mut self) {
         if let Self::Semaphore(sem) = self {
@@ -330,12 +332,13 @@ pub struct TopicMonitor {
     status_: AtomicU8,
 
     /// Binary semaphore used to communicate changes.
-    /// If status_ is STATUS_NEEDS_WAKEUP, then a thread has commited to call wait() on our sema and
+    /// If status_ is STATUS_NEEDS_WAKEUP, then a thread has committed to call wait() on our sema and
     /// this must be balanced by the next call to post(). Note only one thread may wait at a time.
     sema_: BinarySemaphore,
 }
 
 // safety: this is only needed for tests
+#[cfg(test)]
 unsafe impl Sync for TopicMonitor {}
 
 /// The principal topic monitor.
@@ -377,7 +380,7 @@ impl TopicMonitor {
         }
         // Note that if the STATUS_NEEDS_WAKEUP bit is set, no other bits must be set.
         assert!(
-            ((oldstatus == STATUS_NEEDS_WAKEUP) == ((oldstatus & STATUS_NEEDS_WAKEUP) != 0)),
+            (oldstatus == STATUS_NEEDS_WAKEUP) == ((oldstatus & STATUS_NEEDS_WAKEUP) != 0),
             "If STATUS_NEEDS_WAKEUP is set no other bits should be set"
         );
 
