@@ -2,7 +2,7 @@
 
 use crate::common::{get_ellipsis_char, get_ellipsis_str};
 use crate::expand::INTERNAL_SEPARATOR;
-use crate::fallback::{fish_wcwidth, wcscasecmp};
+use crate::fallback::{fish_wcwidth, wcscasecmp, wcscasecmp_fuzzy};
 use crate::flog::FLOGF;
 use crate::libc::MB_CUR_MAX;
 use crate::wchar::{decode_byte_from_char, prelude::*};
@@ -28,6 +28,18 @@ pub fn string_prefixes_string_case_insensitive(proposed_prefix: &wstr, value: &w
     prefix_size <= value.len() && wcscasecmp(&value[..prefix_size], proposed_prefix).is_eq()
 }
 
+pub fn string_prefixes_string_maybe_case_insensitive(
+    icase: bool,
+    proposed_prefix: &wstr,
+    value: &wstr,
+) -> bool {
+    (if icase {
+        string_prefixes_string_case_insensitive
+    } else {
+        string_prefixes_string
+    })(proposed_prefix, value)
+}
+
 /// Test if a string is a suffix of another.
 pub fn string_suffixes_string_case_insensitive(proposed_suffix: &wstr, value: &wstr) -> bool {
     let suffix_size = proposed_suffix.len();
@@ -83,21 +95,25 @@ pub fn ifind(haystack: &wstr, needle: &wstr, fuzzy: bool /* = false */) -> Optio
         .as_char_slice()
         .windows(needle.len())
         .position(|window| {
-            for (l, r) in window.iter().zip(needle.chars()) {
-                // In fuzzy matching treat treat `-` and `_` as equal (#3584).
-                if fuzzy && ['-', '_'].contains(l) && ['-', '_'].contains(&r) {
-                    continue;
-                }
-                // TODO Decide what to do for different lengths.
-                let l = l.to_lowercase();
-                let r = r.to_lowercase();
-                for (l, r) in l.zip(r) {
-                    if l != r {
-                        return false;
-                    }
+            // In fuzzy matching treat treat `-` and `_` as equal (#3584).
+            fn fuzzy_canonicalize(c: char) -> char {
+                if c == '_' {
+                    '-'
+                } else {
+                    c
                 }
             }
-            true
+
+            wcscasecmp_fuzzy(
+                wstr::from_char_slice(window),
+                needle,
+                if fuzzy {
+                    fuzzy_canonicalize
+                } else {
+                    std::convert::identity
+                },
+            )
+            .is_eq()
         })
 }
 
@@ -319,7 +335,7 @@ pub fn wcs2string_callback(input: &wstr, mut func: impl FnMut(&[u8]) -> bool) ->
             let len = unsafe {
                 wcrtomb(
                     std::ptr::addr_of_mut!(converted[0]).cast(),
-                    c as libc::wchar_t,
+                    c as u32,
                     &mut state,
                 )
             };
@@ -389,7 +405,7 @@ pub fn split_string_tok<'val>(
 }
 
 /// Joins strings with a separator.
-/// This supports both &[&wstr] and &[&WString].
+/// This supports both <code>&[&[wstr]]</code> and <code>&[&[WString]]</code>
 pub fn join_strings<S: AsRef<wstr>>(strs: &[S], sep: char) -> WString {
     if strs.is_empty() {
         return WString::new();
